<?php

namespace Drupal\responsive_image_style_builder\Services;

use Drupal\image\Entity\ImageStyle;
use Drupal\Core\Language\LanguageInterface;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Component\DependencyInjection\Container;
use Drupal\responsive_image\Entity\ResponsiveImageStyle;

class ResponsiveImageStyleBuilder
{
  /**
   * Constructs a new ResponsiveImageStyleBuilder instance.
   *
   * @param EntityTypeManagerInterface $entityTypeManager
   *   The entity type manager service used for loading and managing entity types.
   */
  public function __construct(
    protected EntityTypeManagerInterface $entityTypeManager
  ) {}

  /**
   * {@inheritdoc}
   */
  public static function create(Container $container): self
  {
    return new static(
      $container->get('entity_type.manager')
    );
  }

  /**
   * Retrieves a list of available image styles as options for selection.
   *
   * @return array
   *   An associative array of image style options, with image style IDs as keys
   *   and their labels as values. Includes a 'None' option as the first entry.
   */
  public function getImageStyleOptions(): array
  {
    $options = ['' => t('None')];
    $storage = $this->entityTypeManager->getStorage('image_style');
    $imageStyles = $storage->loadMultiple();

    foreach ($imageStyles as $imageStyleId => $imageStyle) {
      $options[$imageStyleId] = $imageStyle->label();
    }

    return $options;
  }

  /**
   * Creates a fallback image style for a responsive image style.
   *
   * Generates a default image style with a machine name derived from the
   * responsive image style's ID, using the same label as the original style.
   *
   * @param ResponsiveImageStyle $responsiveImageStyle
   *   The responsive image style for which to create a fallback.
   *
   * @return ImageStyle
   *   The created or existing fallback image style.
   */
  public function createFallbackImageStyle(ResponsiveImageStyle $responsiveImageStyle): ImageStyle
  {
    $responsiveImageStyleId = sprintf('%s_default', $responsiveImageStyle->id());
    return $this->createImageStyle(['id' => $responsiveImageStyleId, 'label' => $responsiveImageStyle->label()]);
  }

  /**
   * Fetches an image style by given properties.
   * In case there's a new image style to be created, the label is automatically set.
   *
   * @param array $imageStyle
   * @return ImageStyle
   */
  private function createImageStyle(array|string $imageStyleConfig): ImageStyle
  {
    if (is_string($imageStyleConfig)) {
      $imageStyleConfig = [
        'id' => $this->getMachineName($imageStyleConfig),
        'label' => $imageStyleConfig
      ];
    }

    $storage = $this->entityTypeManager->getStorage('image_style');
    $imageStyle = $storage->load($imageStyleConfig['id']);

    if (!$imageStyle) {
      $imageStyle = ImageStyle::create([
        'name' => $imageStyleConfig['id'],
        'label' => $imageStyleConfig['label'] ?? $imageStyleConfig['id']
      ]);

      foreach ($this->getDefaultEffects($imageStyle->id()) as $imageEffect) {
        $imageStyle->addImageEffect($imageEffect);
      }
      $imageStyle->save();
    }
    return $imageStyle;
  }

  /**
   * Retrieves an image style by its identifier.
   *
   * @param string $imageStyleId The unique identifier of the image style to load
   * @return ImageStyle|null The loaded image style, or null if not found
   */
  public function getImageStyle(string $imageStyleId): ?ImageStyle
  {
    $storage = $this->entityTypeManager->getStorage('image_style');
    return $storage->load($imageStyleId);
  }

  /**
   * Creates a derivative image style based on the provided configuration.
   *
   * @param array $imageStyleConfig Configuration details for the derivative image style
   * @return ImageStyle The created or loaded image style
   */
  public function createDerivativeImageStyle(array $imageStyleConfig): ImageStyle
  {
    return $this->createImageStyle($imageStyleConfig);
  }

  /**
   * Retrieves derivative image styles for a given parent image style.
   *
   * Finds all image styles that are derivatives of the specified parent style,
   * identified by sharing a base name and ending with a multiplier (e.g., 2x).
   *
   * @param string $parentImageStyleId The ID of the parent image style
   * @return array An associative array of derivative image styles, keyed by their IDs
   */
  public function getDerivativeImageStyles(string $parentImageStyleId): array
  {
    // Get base name from parent (removing _default)
    $baseStyleName = str_replace('_default', '', $parentImageStyleId);

    $derivativeImageStyles = [];
    $styles = $this->entityTypeManager
      ->getStorage('image_style')
      ->loadMultiple();
    foreach ($styles as $imageStyleId => $imageStyle) {
        // Skip the parent
        if ($imageStyleId === $parentImageStyleId) {
            continue;
        }

        // Check if style starts with the base name and ends with multiplier
        if (strpos($imageStyleId, $baseStyleName . '_') === 0 &&
            preg_match('/_[0-9]+x$/', $imageStyleId)) {
            $derivativeImageStyles[$imageStyleId] = $imageStyle;
        }
    }

    return $derivativeImageStyles;
  }

  /**
   * Extracts the base image style ID by removing multiplier and derivative suffixes.
   *
   * @param string $imageStyleId The full image style identifier
   * @return string The base image style ID without multiplier or derivative suffixes
   */
  private function getImageStyleBaseId(string $imageStyleId): string
  {
    $stripMultiplier = substr($imageStyleId, 0, strrpos($imageStyleId, '_'));
    $imageStyleBaseId = substr($stripMultiplier, 0, strrpos($stripMultiplier, '_'));

    return $imageStyleBaseId;
  }

  /**
   * Returns default effects, or if an id is given, the effects of the parent.
   *
   * @param string $imageStyleId
   * @return array
   */
  private function getDefaultEffects(string $imageStyleId = ''): array
  {
    $i = 0;
    $effects = [];

    // Check if there's a "parent" image style
    $baseImageStyleId = $this->getImageStyleBaseId($imageStyleId);

    if (!empty($baseImageStyleId)) {
      /** @var ImageStyle $parentImageStyle */
      $parentImageStyle = \Drupal::service('risb.builder')->getImageStyle($baseImageStyleId . '_default');

      if ($parentImageStyle) {
        foreach ($parentImageStyle->getEffects() as $effect) {
          $effects[] = $effect->getConfiguration();
        }
        return $effects;
      }
    }

    // @todo Make this variable overrideable
    $configurations = [
      'image_scale_and_crop' => [
        'width' => 100,
        'height' => 100
      ],
      'image_convert' => [
        'format' => 'webp',
        'quality' => 100,
      ],
    ];
    \Drupal::moduleHandler()->alter('risb_default_image_style_effects', $configurations);

    foreach ($configurations as $configurationId => $data) {
      $configuration = [
        'uuid' => NULL,
        'id' => $configurationId,
        'weight' => $i * 10,
        'data' => $data,
      ];
      $effect = \Drupal::service('plugin.manager.image.effect')->createInstance($configuration['id'], $configuration);
      $effects[] = $effect->getConfiguration();
      $i += 1;
    }

    return $effects;
  }

  /**
   * Converts a string to a valid machine name.
   *
   * Transliterates the input string, converts to lowercase, and replaces
   * non-alphanumeric characters with underscores.
   *
   * @param string $string
   *   The input string to convert to a machine name.
   *
   * @return string
   *   A valid machine name derived from the input string.
   */
  private function getMachineName($string): string
  {
    $transliterated = \Drupal::transliteration()->transliterate($string, LanguageInterface::LANGCODE_DEFAULT, '_');
    $transliterated = mb_strtolower($transliterated);
    $transliterated = preg_replace('@[^a-z0-9_.]+@', '_', $transliterated);
    return $transliterated;
  }

  /**
   * Determines if an image effect requires a configuration step.
   *
   * @param string $effectId
   *   The ID of the image effect to check.
   *
   * @return bool
   *   TRUE if the effect requires configuration, FALSE otherwise.
   */
  public function effectHasConfigurationStep(string $effectId): bool
  {
    // @todo Add a hook to extend this list
    $imageStylesWithoutConfiguration = ['image_desaturate'];

    if(in_array($effectId, $imageStylesWithoutConfiguration)) {
      return FALSE;
    }
    return TRUE;
  }
}
